"""
MobileNet v2, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/mobilenetv2.py

Paper "MobileNetV2: Inverted Residuals and Linear Bottlenecks", https://arxiv.org/abs/1801.04381
"""

# Reference license: BSD 3-Clause

from collections.abc import Callable
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.core.net.base import BaseNet
from birder.core.net.base import make_divisible


class InvertedResidual(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        expansion_factor: float,
        shortcut: bool,
        activation_layer: Callable[..., nn.Module] = nn.ReLU6,
    ) -> None:
        super().__init__()
        num_expfilter = int(round(in_channels * expansion_factor))

        self.shortcut = shortcut
        self.block = nn.Sequential(
            Conv2dNormActivation(
                in_channels,
                num_expfilter,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=activation_layer,
            ),
            Conv2dNormActivation(
                num_expfilter,
                num_expfilter,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=num_expfilter,
                bias=False,
                activation_layer=activation_layer,
            ),
            Conv2dNormActivation(
                num_expfilter,
                out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=None,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.shortcut is True:
            return x + self.block(x)

        return self.block(x)


# pylint: disable=invalid-name
class MobileNet_v2(BaseNet):
    default_size = 224

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        net_param: Optional[float] = None,
        size: Optional[int] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, net_param, size)
        alpha = net_param
        alpha_values = [0.25, 0.50, 0.75, 1.0, 1.25, 1.5, 1.75, 2.0]
        assert alpha in alpha_values, f"alpha = {alpha} not supported"

        # t - expansion factor
        # c - num_filters (channels)
        # n - number of repetitions
        # s - stride
        inverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        base = make_divisible(32 * alpha, 8)
        self.stem = Conv2dNormActivation(
            self.input_channels,
            base,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            bias=False,
            activation_layer=nn.ReLU6,
        )

        layers = []
        for t, c, n, s in inverted_residual_setting:
            c = make_divisible(c * alpha, 8)
            layers.append(
                InvertedResidual(
                    base,
                    c,
                    kernel_size=(3, 3),
                    stride=(s, s),
                    padding=(1, 1),
                    expansion_factor=t,
                    shortcut=False,
                )
            )
            for _ in range(1, n):
                layers.append(
                    InvertedResidual(
                        c,
                        c,
                        kernel_size=(3, 3),
                        stride=(1, 1),
                        padding=(1, 1),
                        expansion_factor=t,
                        shortcut=True,
                    )
                )

            base = c

        last_channels = make_divisible(1280 * max(1.0, alpha), 8)
        layers.append(
            Conv2dNormActivation(
                c,
                last_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=nn.ReLU6,
            )
        )

        self.body = nn.Sequential(*layers)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.embedding_size = last_channels
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d) is True:
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)) is True:
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear) is True:
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        x = self.body(x)
        return self.features(x)
